@@ -12,27 +12,27 @@ import re
 import signal
 import string
 import sys
+import traceback
 
-from collections import defaultdict, OrderedDict
-from concurrent.futures import ThreadPoolExecutor, as_completed
-from datetime import datetime
+from collections import deque
 from functools import partial
 from multiprocessing import Pool
 from os import makedirs, path, walk
 from time import sleep
 
-from courlan import extract_domain, validate_url
+from courlan import get_host_and_path, is_navigation_page, validate_url
 
 from .core import extract
+from .downloads import add_to_compressed_dict, buffered_downloads, load_download_buffer
 from .filters import content_fingerprint
-from .settings import (use_config, DOWNLOAD_THREADS, FILENAME_LEN, FILE_PROCESSING_CORES,
-                       MIN_FILE_SIZE, MAX_FILE_SIZE, MAX_FILES_PER_DIRECTORY,
-                       PROCESSING_TIMEOUT)
-from .utils import fetch_url, HOSTINFO
+from .settings import (use_config, FILENAME_LEN,
+                       FILE_PROCESSING_CORES, MAX_FILES_PER_DIRECTORY)
+from .spider import get_crawl_delay, init_crawl, process_response
 
 
 LOGGER = logging.getLogger(__name__)
 random.seed(345)  # make generated file names reproducible
+CHAR_CLASS = string.ascii_letters + string.digits
 
 
 # try signal https://stackoverflow.com/questions/492519/timeout-on-a-function-call
@@ -41,54 +41,25 @@ def handler(signum, frame):
     raise Exception('unusual file processing time, aborting')
 
 
-def load_input_dict(filename, blacklist):
-    '''Read input list of URLs to process and build domain-aware processing dictionary'''
-    inputdict = defaultdict(list)
-    try:
-        # optional: errors='strict', buffering=1
-        with open(filename, mode='r', encoding='utf-8') as inputfile:
-            for line in inputfile:
-                # control input validity
-                url_match = re.match(r'https?://[^\s]+', line)
-                if url_match:
-                    url = url_match.group(0)
-                    # validation
-                    if validate_url(url)[0] is False:
-                        LOGGER.warning('Invalid URL, discarding line: %s', line)
-                        continue
-                    # control blacklist
-                    if blacklist:
-                        if re.sub(r'^https?://', '', url) in blacklist:
-                            continue
-                    # segment URL and add to domain dictionary
-                    try:
-                        _, hostinfo, urlpath = HOSTINFO.split(url)
-                        inputdict[hostinfo].append(urlpath)
-                    except ValueError:
-                        LOGGER.warning('Could not parse URL, discarding line: %s', line)
-                else:
-                    LOGGER.warning('Not an URL, discarding line: %s', line)
-    except UnicodeDecodeError:
-        sys.exit('ERROR: system, file type or buffer encoding')
-    # deduplicate
-    for hostname in inputdict:
-        inputdict[hostname] = list(OrderedDict.fromkeys(inputdict[hostname]))
-    return inputdict
-
-
-def load_input_urls(filename):
-    '''Read list of URLs to process'''
-    input_urls = []
-    try:
-        with open(filename, mode='r', encoding='utf-8') as inputfile:
-            for line in inputfile:
-                url_match = re.match(r'https?://[^\s]+', line)
-                try:
-                    input_urls.append(url_match.group(0))
-                except AttributeError:
-                    LOGGER.warning('Not an URL, discarding line: %s', line)
-    except UnicodeDecodeError:
-        sys.exit('ERROR: system, file type or buffer encoding')
+def load_input_urls(args):
+    '''Read list of URLs to process or derive one from command-line arguments'''
+    if args.inputfile:
+        input_urls = []
+        try:
+            # optional: errors='strict', buffering=1
+            with open(args.inputfile, mode='r', encoding='utf-8') as inputfile:
+                for line in inputfile:
+                    url_match = re.match(r'https?://[^\s]+', line)
+                    if url_match:
+                        input_urls.append(url_match.group(0))
+        except UnicodeDecodeError:
+            sys.exit('ERROR: system, file type or buffer encoding')
+    elif args.crawl:
+        input_urls = [args.crawl]
+    elif args.feed:
+        input_urls = [args.feed]
+    elif args.sitemap:
+        input_urls = [args.sitemap]
     return input_urls
 
 
@@ -103,34 +74,12 @@ def load_blacklist(filename):
     return blacklist
 
 
-def convert_inputlist(blacklist, inputlist, url_filter=None, inputdict=None):
-    '''Add input URls to domain-aware processing dictionary'''
-    # control
-    if inputdict is None:
-        inputdict = defaultdict(list)
-    # filter
-    if blacklist:
-        inputlist = [u for u in inputlist if re.sub(r'https?://', '', u) not in blacklist]
-    if url_filter:
-        filtered_list = []
-        while inputlist:
-            u = inputlist.pop()
-            for f in url_filter:
-                if f in u:
-                    filtered_list.append(u)
-                    break
-        inputlist = filtered_list
-    # validate
-    inputlist = [u for u in inputlist if validate_url(u)[0] is True]
-    # deduplicate
-    for url in list(OrderedDict.fromkeys(inputlist)):
-        # segment URL and add to domain dictionary
-        try:
-            _, hostinfo, urlpath = HOSTINFO.split(url)
-            inputdict[hostinfo].append(urlpath)
-        except ValueError:
-            LOGGER.warning('Could not parse URL, discarding: %s', url)
-    return inputdict
+def load_input_dict(args):
+    '''Read input list of URLs to process and
+       build a domain-aware dictionary'''
+    inputlist = load_input_urls(args)
+    # deduplicate, filter and and convert to dict
+    return add_to_compressed_dict(inputlist, args.blacklist) # args.filter
 
 
 def check_outputdir_status(directory):
@@ -160,11 +109,10 @@ def determine_counter_dir(dirname, counter):
 
 def get_writable_path(destdir, extension):
     '''Find a writable path and return it along with its random file name'''
-    charclass = string.ascii_letters + string.digits
-    filename = ''.join(random.choice(charclass) for _ in range(FILENAME_LEN))
+    filename = ''.join(random.choice(CHAR_CLASS) for _ in range(FILENAME_LEN))
     output_path = path.join(destdir, filename + extension)
     while path.exists(output_path):
-        filename = ''.join(random.choice(charclass) for _ in range(FILENAME_LEN))
+        filename = ''.join(random.choice(CHAR_CLASS) for _ in range(FILENAME_LEN))
         output_path = path.join(destdir, filename + extension)
     return output_path, filename
 
@@ -173,11 +121,11 @@ def determine_output_path(args, orig_filename, content, counter=None, new_filena
     '''Pick a directory based on selected options and a file name based on output type'''
     # determine extension
     extension = '.txt'
-    if args.xml or args.xmltei or args.output_format == 'xml':
+    if args.output_format in ('xml', 'xmltei'):
         extension = '.xml'
-    elif args.csv or args.output_format == 'csv':
+    elif args.output_format == 'csv':
         extension = '.csv'
-    elif args.json or args.output_format == 'json':
+    elif args.output_format == 'json':
         extension = '.json'
     # use cryptographic hash on file contents to define name
     if args.hash_as_name is True:
@@ -250,70 +198,81 @@ def process_result(htmlstring, args, url, counter, config):
         fileslug = None
     # process
     result = examine(htmlstring, args, url=url, config=config)
-    write_result(result, args, orig_filename=None, counter=None, new_filename=fileslug)
+    write_result(result, args, orig_filename=fileslug, counter=counter, new_filename=fileslug)
     # increment written file counter
     if counter is not None and result is not None:
         counter += 1
     return counter
 
 
-def draw_backoff_url(domain_dict, backoff_dict, sleeptime, i):
-    '''Select a random URL from the domains pool and apply backoff rule'''
-    host = random.choice(list(domain_dict))
-    domain = extract_domain(host)
-    # safeguard
-    if domain in backoff_dict and \
-        (datetime.now() - backoff_dict[domain]).total_seconds() < sleeptime:
-        i += 1
-        if i >= len(domain_dict)*3:
-            LOGGER.debug('spacing request for domain name %s', domain)
-            sleep(sleeptime)
-            i = 0
-    # draw URL
-    url = host + domain_dict[host].pop()
-    # clean registries
-    if not domain_dict[host]:
-        del domain_dict[host]
-        try:
-            del backoff_dict[domain]
-        except KeyError:
-            pass
-    # register backoff
-    else:
-        backoff_dict[domain] = datetime.now()
-    return url, domain_dict, backoff_dict, i
-
-
-def download_queue_processing(domain_dict, args, sleeptime, counter, config):
+def download_queue_processing(domain_dict, args, counter, config):
     '''Implement a download queue consumer, single- or multi-threaded'''
-    i, backoff_dict, errors = 0, dict(), []
+    sleep_time = config.getfloat('DEFAULT', 'SLEEP_TIME')
+    backoff_dict, errors = dict(), []
     while domain_dict:
-        download_threads = args.parallel or DOWNLOAD_THREADS
-        # the remaining list is too small, process it differently
-        if len(domain_dict) < download_threads or \
-           len({x for v in domain_dict.values() for x in v}) < download_threads:
-            download_threads, i = 1, 3
-        # populate buffer
-        bufferlist = []
-        while len(bufferlist) < download_threads:
-            url, domain_dict, backoff_dict, i = draw_backoff_url(domain_dict, backoff_dict, sleeptime, i)
-            bufferlist.append(url)
-        # start several threads
-        with ThreadPoolExecutor(max_workers=download_threads) as executor:
-            future_to_url = {executor.submit(fetch_url, url): url for url in bufferlist}
-            for future in as_completed(future_to_url):
-                url = future_to_url[future]
-                # handle result
-                if future.result() is not None:
-                    counter = process_result(future.result(), args, url, counter, config)
-                else:
-                    LOGGER.debug('No result for URL: %s', url)
-                    if args.archived is True:
-                        errors.append(url)
+        bufferlist, download_threads, domain_dict, backoff_dict = load_download_buffer(domain_dict, backoff_dict, sleep_time, threads=args.parallel)
+        # process downloads
+        for url, result in buffered_downloads(bufferlist, download_threads):
+            # handle result
+            if result is not None and result != '':
+                counter = process_result(result, args, url, counter, config)
+            else:
+                LOGGER.debug('No result for URL: %s', url)
+                if args.archived is True:
+                    errors.append(url)
     return errors, counter
 
 
-def url_processing_pipeline(args, inputdict, sleeptime):
+def cli_crawler(args, n=30):
+    '''Start a focused crawler which downloads a fixed number of URLs within a website
+       and prints the links found in the process'''
+    config = use_config(filename=args.config_file)
+    sleep_time = config.getfloat('DEFAULT', 'SLEEP_TIME')
+    counter, crawlinfo, backoff_dict = None, dict(), dict()
+    # load input URLs
+    domain_dict = load_input_dict(args)
+    # load crawl data
+    for website in domain_dict:
+        homepage = website + domain_dict[website].popleft()
+        crawlinfo[website] = dict()
+        domain_dict[website], crawlinfo[website]['known'], crawlinfo[website]['base'], crawlinfo[website]['count'], crawlinfo[website]['rules'] = init_crawl(homepage, None, set(), language=args.target_language, shortform=True)
+        # update info
+        # TODO: register changes?
+        # if base_url != website:
+        # ...
+    # iterate until the threshold is reached
+    while domain_dict:
+        bufferlist, download_threads, domain_dict, backoff_dict = load_download_buffer(domain_dict, backoff_dict, sleep_time, threads=args.parallel)
+        # start several threads
+        for url, result in buffered_downloads(bufferlist, download_threads, decode=False):
+            website, _ = get_host_and_path(url)
+            crawlinfo[website]['count'] += 1
+            # handle result
+            if result is not None and result != '':
+                domain_dict[website], crawlinfo[website]['known'], htmlstring = process_response(result, domain_dict[website], crawlinfo[website]['known'], crawlinfo[website]['base'], args.target_language, shortform=True, rules=crawlinfo[website]['rules'])
+                # only store content pages, not navigation
+                if not is_navigation_page(url): # + response.geturl()
+                    if args.list:
+                        write_result(url, args)
+                    else:
+                        counter = process_result(htmlstring, args, url, counter, config)
+                # just in case a crawl delay is specified in robots.txt
+                sleep(get_crawl_delay(crawlinfo[website]['rules']))
+                #else:
+                #    LOGGER.debug('No result for URL: %s', url)
+                #    if args.archived is True:
+                #        errors.append(url)
+        # early exit if maximum count is reached
+        if any(i >= n for i in [crawlinfo[site]['count'] for site in crawlinfo]):
+            break
+    # print results
+    for website in sorted(domain_dict):
+        for urlpath in sorted(domain_dict[website]):
+            sys.stdout.write(website + urlpath +'\n')
+    #return todo, known_links
+
+
+def url_processing_pipeline(args, inputdict):
     '''Aggregated functions to show a list and download and process an input list'''
     # print list without further processing
     if args.list:
@@ -331,14 +290,14 @@ def url_processing_pipeline(args, inputdict, sleeptime):
             counter = 0
             break
     # download strategy
-    errors, counter = download_queue_processing(inputdict, args, sleeptime, counter, config)
+    errors, counter = download_queue_processing(inputdict, args, counter, config)
     LOGGER.debug('%s URLs could not be found', len(errors))
     # option to retry
     if args.archived is True:
         inputdict = dict()
-        inputdict['https://web.archive.org'] = ['/web/20/' + e for e in errors]
+        inputdict['https://web.archive.org'] = deque(['/web/20/' + e for e in errors])
         if len(inputdict['https://web.archive.org']) > 0:
-            archived_errors, _ = download_queue_processing(inputdict, args, sleeptime, counter, config)
+            archived_errors, _ = download_queue_processing(inputdict, args, counter, config)
             LOGGER.debug('%s archived URLs out of %s could not be found', len(archived_errors), len(errors))
 
 
@@ -369,31 +328,32 @@ def file_processing_pipeline(args):
 def examine(htmlstring, args, url=None, config=None):
     """Generic safeguards and triggers"""
     result = None
+    if config is None:
+        config = use_config(filename=args.config_file)
     # safety check
     if htmlstring is None:
         sys.stderr.write('ERROR: empty document\n')
-    elif len(htmlstring) > MAX_FILE_SIZE:
+    elif len(htmlstring) > config.getint('DEFAULT', 'MAX_FILE_SIZE'):
         sys.stderr.write('ERROR: file too large\n')
-    elif len(htmlstring) < MIN_FILE_SIZE:
+    elif len(htmlstring) < config.getint('DEFAULT', 'MIN_FILE_SIZE'):
         sys.stderr.write('ERROR: file too small\n')
     # proceed
     else:
         # put timeout signal in place
-        if args.timeout is True:
-            signal.signal(signal.SIGALRM, handler)
-            signal.alarm(PROCESSING_TIMEOUT)
+        signal.signal(signal.SIGALRM, handler)
+        signal.alarm(config.getint('DEFAULT', 'EXTRACTION_TIMEOUT'))
         try:
             result = extract(htmlstring, url=url, no_fallback=args.fast,
                              include_comments=args.nocomments, include_tables=args.notables,
-                             include_formatting=args.formatting,
+                             include_formatting=args.formatting, include_links=args.links,
+                             include_images=args.images, only_with_metadata=args.only_with_metadata,
                              with_metadata=args.with_metadata,
-                             output_format=args.output_format, tei_validation=args.validate,
+                             output_format=args.output_format, tei_validation=args.validate_tei,
                              target_language=args.target_language, deduplicate=args.deduplicate,
-                             settingsfile=args.config_file, config=config)
+                             config=config) # settingsfile=args.config_file,
         # ugly but efficient
         except Exception as err:
-            sys.stderr.write('ERROR: ' + str(err) + '\nDetails: ' + str(sys.exc_info()[0]) + '\n')
+            sys.stderr.write('ERROR: ' + str(err) + '\n' + traceback.format_exc() + '\n')
         # deactivate
-        if args.timeout is True:
-            signal.alarm(0)
+        signal.alarm(0)
     return result
