@@ -10,6 +10,7 @@ Module bundling all functions needed to extract the text in a webpage.
 # standard
 import logging
 import re # import regex as re
+import warnings
 
 from collections import OrderedDict
 from copy import deepcopy
@@ -20,29 +21,30 @@ from lxml import etree, html
 from .external import justext_rescue, sanitize_tree, SANITIZED_XPATH, try_readability
 from .filters import (check_html_lang, content_fingerprint, duplicate_test,
                      language_filter, text_chars_test)
-from .htmlprocessing import (convert_tags, discard_unwanted,
-                             discard_unwanted_comments, handle_textnode,
+from .htmlprocessing import (convert_tags, handle_textnode,
                              link_density_test, link_density_test_tables,
-                             process_node, tree_cleaning)
+                             process_node, prune_unwanted_nodes, tree_cleaning)
 from .metadata import extract_metadata, METADATA_LIST
 from .settings import use_config, DEFAULT_CONFIG, TAG_CATALOG
 from .utils import load_html, trim, txttocsv, is_image_file
 from .xml import (build_json_output, build_xml_output, build_tei_output,
                   control_xml_output, xmltotxt)
-from .xpaths import BODY_XPATH, COMMENTS_XPATH
+from .xpaths import BODY_XPATH, COMMENTS_XPATH, COMMENTS_DISCARD_XPATH, DISCARD_XPATH
+#, REMOVE_COMMENTS_XPATH
 
 
 LOGGER = logging.getLogger(__name__)
 
 
-def handle_titles(element):
+def handle_titles(element, dedupbool, config):
     '''Process head elements (titles)'''
     # maybe needs attention
     if element.tail and re.search(r'\w', element.tail):
         LOGGER.debug('tail in title, stripping: %s', element.tail)
     element.tail = None
-    title = process_node(element)
-    if title is not None and title.text and re.search(r'\w', title.text):
+    title = process_node(element, dedupbool, config)
+    # and re.search(r'\w', title.text)
+    if title is not None and text_chars_test(title.text) is True:
         return title
     return None
 
@@ -64,10 +66,14 @@ def handle_formatting(element):
 def handle_lists(element, dedupbool, config):
     '''Process lists elements'''
     processed_element = etree.Element(element.tag)
+    if element.text is not None:
+        processed_element.text = element.text
+    #if element.tail is not None:
+    #    processed_element.tail = element.text
     for child in element.iter('item'):
         newchildelem = etree.Element('item')
         if len(child) == 0:
-            processed_child = process_node(child)
+            processed_child = process_node(child, dedupbool, config)
             if processed_child is not None:
                 newchildelem.text, newchildelem.tail = processed_child.text, processed_child.tail
                 processed_element.append(newchildelem)
@@ -84,29 +90,26 @@ def handle_lists(element, dedupbool, config):
         if newchildelem.text or len(newchildelem) > 0:
             processed_element.append(newchildelem)
         child.tag = 'done'
-    # avoid double tags??
-    if len(processed_element) > 0:  # if it has children
-        # test if it has text
-        if text_chars_test(''.join(processed_element.itertext())) is True:
-            return processed_element
+    # test if it has children and text. Avoid double tags??
+    if len(processed_element) > 0 and \
+        text_chars_test(''.join(processed_element.itertext())) is True:
+        return processed_element
     return None
 
 
-def handle_quotes(element):
+def handle_quotes(element, dedupbool, config):
     '''Process quotes elements'''
     processed_element = etree.Element(element.tag)
     for child in element.iter():
-        processed_child = process_node(child) # handle_textnode(child, comments_fix=True)
+        processed_child = process_node(child, dedupbool, config) # handle_textnode(child, comments_fix=True)
         if processed_child is not None:
             newsub = etree.SubElement(processed_element, child.tag)
             newsub.text, newsub.tail = processed_child.text, processed_child.tail
         child.tag = 'done'
-    if len(processed_element) > 0:
+    if len(processed_element) > 0 and \
+        text_chars_test(''.join(processed_element.itertext())) is True:
         # avoid double/nested tags
         etree.strip_tags(processed_element, 'quote')
-        # test if it has text
-        # teststring = ''.join(processed_element.itertext())
-        # if len(teststring) > 0 and re.search(r'[p{L}]', teststring):
         return processed_element
     return None
 
@@ -119,7 +122,7 @@ def handle_other_elements(element, potential_tags, dedupbool, config):
         return None
     if element.tag == 'div':
         processed_element = handle_textnode(element, comments_fix=False, deduplicate=dedupbool, config=config)
-        if processed_element is not None:
+        if processed_element is not None and text_chars_test(processed_element.text) is True:
             processed_element.attrib.clear()
             # small div-correction # could be moved elsewhere
             if processed_element.tag == 'div':
@@ -135,10 +138,10 @@ def handle_paragraphs(element, potential_tags, dedupbool, config):
     '''Process paragraphs (p) elements along with their children,
        trim and clean the content'''
     element.attrib.clear()
-    #etree.strip_tags(element, 'p')  # change in precision
+    # etree.strip_tags(element, 'p') # change in precision due to spaces?
     # no children
     if len(element) == 0:
-        processed_element = process_node(element)  # handle_textnode(element, comments_fix=False)
+        processed_element = process_node(element, dedupbool, config)
         if processed_element is not None:
             return processed_element
         return None
@@ -150,41 +153,51 @@ def handle_paragraphs(element, potential_tags, dedupbool, config):
             continue
         processed_child = handle_textnode(child, comments_fix=False, deduplicate=dedupbool, config=config)
         if processed_child is not None:
-            # needing attention!
-            if child.tag == 'p':
-                LOGGER.debug('extra p within p: %s %s %s', child.tag, child.text, child.tail)
+            # todo: needing attention!
+            if processed_child.tag == 'p':
+                LOGGER.debug('extra p within p: %s %s %s', processed_child.tag, processed_child.text, processed_child.tail)
                 if processed_element.text:
-                    processed_element.text += ' ' + trim(child.text)
+                    processed_element.text += ' ' + processed_child.text
                 else:
-                    processed_element.text = trim(child.text)
+                    processed_element.text = processed_child.text
                 continue
             newsub = etree.Element(child.tag)
             # handle formatting
-            if child.tag == 'hi':
+            if processed_child.tag in ('hi', 'ref'):
                 # check depth and clean
-                if len(child) > 0:
-                    for item in child:  # children are lists
+                if len(processed_child) > 0:
+                    for item in processed_child:  # children are lists
                         if text_chars_test(item.text) is True:
                             item.text = ' ' + item.text
-                        etree.strip_tags(child, item.tag)
-                newsub.set('rend', child.get('rend'))
+                        etree.strip_tags(processed_child, item.tag)
+                if child.tag == 'hi':
+                    newsub.set('rend', child.get('rend'))
+                elif child.tag == 'ref':
+                    if child.get('target') is not None:
+                        newsub.set('target', child.get('target'))
+                    # to be removed after thorough testing
+                    elif child.get('href') is not None:
+                        newsub.set('target', child.get('href'))
+                        #del processed_child.attrib['href']
             # handle line breaks
-            elif child.tag == 'lb':
-                try:
-                    processed_child.tail = process_node(child).tail
-                except AttributeError:  # no text
-                    pass
+            #elif processed_child.tag == 'lb':
+            #    try:
+            #        processed_child.tail = process_node(child, dedupbool, config).tail
+            #    except AttributeError:  # no text
+            #        pass
             # prepare text
-            if text_chars_test(processed_child.text) is False:
-                processed_child.text = ''
+            # todo: to be moved to handle_textnode()
+            #if text_chars_test(processed_child.text) is False:
+            #    processed_child.text = ''
+            #if text_chars_test(processed_child.tail) is False:
+            #    processed_child.tail = ''
             # if there are already children
-            if len(processed_element) > 0:
-                if text_chars_test(processed_child.tail) is True:
-                    newsub.tail = processed_child.text + processed_child.tail
-                else:
-                    newsub.tail = processed_child.text
-            else:
-                newsub.text, newsub.tail = processed_child.text, processed_child.tail
+            #if len(processed_element) > 0:
+            #    if text_chars_test(processed_child.tail) is True:
+            #        newsub.tail = processed_child.text + processed_child.tail
+            #    else:
+            #        newsub.tail = processed_child.text
+            newsub.text, newsub.tail = processed_child.text, processed_child.tail
             processed_element.append(newsub)
             child.tag = 'done'
     # finish
@@ -197,7 +210,7 @@ def handle_paragraphs(element, potential_tags, dedupbool, config):
     return None
 
 
-def handle_table(table_elem):
+def handle_table(table_elem, dedupbool, config):
     '''Process single table element'''
     newtable = etree.Element('table')
     newrow = etree.Element('row')
@@ -218,15 +231,15 @@ def handle_table(table_elem):
             #    continue
         elif subelement.tag in ('td', 'th'):
             # process
-            processed_cell = process_node(subelement)
-            if processed_cell is None or processed_cell.text is None or not processed_cell.text:
+            processed_cell = process_node(subelement, dedupbool, config)
+            if processed_cell is None or text_chars_test(processed_cell.text) is False:
                 continue
             # define tag
             newsub = etree.SubElement(newrow, 'cell')
             if subelement.tag == 'th':
                 newsub.set('role', 'head')
             newsub.text = processed_cell.text
-            #newrow.append(newsub)
+            #subelement.tag = 'done' ??
         # beware of nested tables
         elif subelement.tag == 'table' and i > 1:
             break
@@ -240,30 +253,45 @@ def handle_table(table_elem):
 
 def handle_image(element):
     '''Process image element'''
-    if not 'data-src' in element.attrib and not 'src' in element.attrib:
-        return None
     # image source
     processed_element = etree.Element(element.tag)
-    if element.get('data-src') is not None and is_image_file(element.get('data-src')):
+    if is_image_file(element.get('data-src')):
         processed_element.set('src', element.get('data-src'))
-    elif element.get('src') is not None and is_image_file(element.get('src')):
+    elif is_image_file(element.get('src')):
         processed_element.set('src', element.get('src'))
+    else:
+        # take the first corresponding attribute
+        for attr in element.attrib:
+            if attr.startswith('data-src') and is_image_file(element.get(attr)):
+                processed_element.set('src', element.get(attr))
+                break
     # additional data
     if element.get('alt') is not None:
         processed_element.set('alt', element.get('alt'))
     if element.get('title') is not None:
         processed_element.set('title', element.get('title'))
+    # don't return an empty element, just None
+    if len(processed_element.attrib) == 0:
+        return None
+    # post-processing: URLs
+    if processed_element.get('src') is not None:
+        url = processed_element.get('src')
+        processed_element.set('src', re.sub(r'^//', 'http://', url))
     return processed_element
 
 
-def recover_wild_paragraphs(tree, result_body, potential_tags=TAG_CATALOG, deduplicate=True, config=None):
-    '''Look for all p-elements, including outside of the determined frame
-       and throughout the document to recover potentially missing text parts'''
-    LOGGER.debug('Taking all p-elements')
+def recover_wild_text(tree, result_body, potential_tags=TAG_CATALOG, deduplicate=True, config=None):
+    '''Look for all previously unconsidered wild elements, including outside of the determined
+       frame and throughout the document to recover potentially missing text parts'''
+    LOGGER.debug('Recovering wild text elements')
     # prune
-    search_tree = discard_unwanted(tree)
-    etree.strip_tags(search_tree, 'a', 'link', 'span')
-    processed_elems = [handle_paragraphs(element, potential_tags, deduplicate, config) for element in search_tree.iter('blockquote', 'code', 'p', 'pre', 'q', 'quote')] # 'head', 'list'
+    search_tree = prune_unwanted_nodes(tree, DISCARD_XPATH)
+    # decide if links are preserved
+    if not 'ref' in potential_tags:
+        etree.strip_tags(search_tree, 'a', 'ref', 'span')
+    else:
+        etree.strip_tags(search_tree, 'span')
+    processed_elems = [handle_textelem(element, potential_tags, deduplicate, config) for element in search_tree.iter('blockquote', 'code', 'div', 'p', 'pre', 'q', 'quote', 'table')]
     result_body.extend(list(filter(None.__ne__, processed_elems)))
     return result_body
 
@@ -274,23 +302,23 @@ def handle_textelem(element, potential_tags, dedupbool, config):
     # bypass: nested elements
     if element.tag == 'list':
         new_element = handle_lists(element, dedupbool, config)
-    elif element.tag == 'quote':   # + 'code'?
-        new_element = handle_quotes(element)
+    elif element.tag == 'quote' or element.tag == 'code':
+        new_element = handle_quotes(element, dedupbool, config)
     elif element.tag == 'head':
-        new_element = handle_titles(element)
+        new_element = handle_titles(element, dedupbool, config)
     elif element.tag == 'p':
         new_element = handle_paragraphs(element, potential_tags, dedupbool, config)
     elif element.tag == 'lb':
         if text_chars_test(element.tail) is True:
-            element = process_node(element)
+            element = process_node(element, dedupbool, config)
             if element is not None:
                 new_element = etree.Element('p')
                 new_element.text = element.tail
-    elif element.tag == 'hi':
-        new_element = handle_formatting(element)
+    elif element.tag in ('hi', 'ref', 'span'):
+        new_element = handle_formatting(element) # process_node(element, dedupbool, config)
     elif element.tag == 'table' and 'table' in potential_tags:
-        new_element = handle_table(element)
-    elif element.tag == 'image':
+        new_element = handle_table(element, dedupbool, config)
+    elif element.tag == 'graphic' and 'graphic' in potential_tags:
         new_element = handle_image(element)
     else:
         # other elements (div, ??, ??)
@@ -318,9 +346,6 @@ def delete_by_link_density(subtree, tagname, backtracking=False):
             if 0 < len(item) < 100 and len(myelems[item]) >= 3:
                 deletions.extend(myelems[item])
                 #print('backtrack:', item)
-            #elif 20 < len(item) < 100 and len(myelems[item]) >= 3:
-            #    deletions.extend(myelems[item])
-            #    print('backtrack 2:', item)
             #else: # and not re.search(r'[?!.]', text):
                 #print(elem.tag, templist)
     for elem in list(OrderedDict.fromkeys(deletions)):
@@ -328,12 +353,19 @@ def delete_by_link_density(subtree, tagname, backtracking=False):
     return subtree
 
 
-def extract_content(tree, include_tables=False, deduplicate=False, config=None):
+def extract_content(tree, include_tables=False, include_images=False, include_links=False, deduplicate=False, config=None):
     '''Find the main content of a page using a set of XPath expressions,
        then extract relevant elements, strip them of unwanted subparts and
        convert them'''
     sure_thing = False
     result_body = etree.Element('body')
+    potential_tags = set(TAG_CATALOG)  # + 'span'?
+    if include_tables is True:
+        potential_tags.add('table')
+    if include_images is True:
+        potential_tags.add('graphic')
+    if include_links is True:
+        potential_tags.add('ref')
     # iterate
     for expr in BODY_XPATH:
         # select tree if the expression has been found
@@ -341,44 +373,57 @@ def extract_content(tree, include_tables=False, deduplicate=False, config=None):
         if not subtree:
             continue
         subtree = subtree[0]
+        #for table_elem in subtree.xpath('//table'):
+        #    print('1', table_elem.text_content())
         # prune
-        subtree = discard_unwanted(subtree)
+        subtree = prune_unwanted_nodes(subtree, DISCARD_XPATH)
+        #for table_elem in subtree.xpath('//table'):
+        #    print('2', table_elem.text_content())
         # remove elements by link density
         subtree = delete_by_link_density(subtree, 'div', backtracking=True)
         subtree = delete_by_link_density(subtree, 'list', backtracking=False)
         subtree = delete_by_link_density(subtree, 'p', backtracking=False)
+        #subtree = delete_by_link_density(subtree, 'head', backtracking=False)
+        # also filter fw/head and quote elements
+        #for table_elem in subtree.xpath('//table'):
+        #    print('3', table_elem.text_content())
         # define iteration strategy
-        potential_tags = set(TAG_CATALOG)  # + 'span'?
-        if include_tables is True:
-            potential_tags.add('table')
+        if 'table' in potential_tags:
             for elem in subtree.iter('table'):
                 if link_density_test_tables(elem) is True:
                     elem.getparent().remove(elem)
+        #for table_elem in subtree.xpath('//table'):
+        #    print('4', table_elem.text_content())
         # skip if empty tree
         if len(subtree) == 0:
             continue
-        # no paragraphs containing text
-        if not subtree.xpath('//p//text()'):
+        # no paragraphs containing text, or not enough
+        ptest = subtree.xpath('//p//text()')
+        if not ptest or len(''.join(ptest)) < config.getint('DEFAULT', 'MIN_EXTRACTED_SIZE')*2:
             potential_tags.add('div')
+            #potential_tags.add('span')
+        if 'ref' not in potential_tags:
+            etree.strip_tags(subtree, 'ref')
+        if 'span' not in potential_tags:
+            etree.strip_tags(subtree, 'span')
         LOGGER.debug(sorted(potential_tags))
-        etree.strip_tags(subtree, 'link', 'span') # 'a',
-        # etree.strip_tags(subtree, 'lb') # BoingBoing-Bug
+        ##etree.strip_tags(subtree, 'lb') # BoingBoing-Bug
         # extract content
         # list(filter(None.__ne__, processed_elems))
         result_body.extend([e for e in
                             [handle_textelem(e, potential_tags, deduplicate, config) for e in subtree.xpath('.//*')]
                             if e is not None])
         # remove trailing titles
-        while len(result_body) > 0 and result_body[-1].tag in ('fw', 'head'): # and result_body[-1].tail is None:
+        while len(result_body) > 0 and result_body[-1].tag in ('fw', 'head'):
             result_body[-1].getparent().remove(result_body[-1])
         # exit the loop if the result has children
-        if len(result_body) > 1: # try to change this to 0 or 2
+        if len(result_body) > 1: # for recall put 0
             LOGGER.debug(expr)
             break
     temp_text = trim(' '.join(result_body.itertext()))
     # try parsing wild <p> elements if nothing found or text too short
     if len(result_body) == 0 or len(temp_text) < config.getint('DEFAULT', 'MIN_EXTRACTED_SIZE'):
-        result_body = recover_wild_paragraphs(tree, result_body, deduplicate=deduplicate, config=config)
+        result_body = recover_wild_text(tree, result_body, potential_tags=potential_tags, deduplicate=deduplicate, config=config)
         temp_text = trim(' '.join(result_body.itertext()))
     else:
         sure_thing = True
@@ -416,8 +461,8 @@ def extract_comments(tree, dedupbool, config):
             continue
         subtree = subtree[0]
         # prune
-        subtree = discard_unwanted_comments(subtree)
-        etree.strip_tags(subtree, 'a', 'link', 'span')
+        subtree = prune_unwanted_nodes(subtree, COMMENTS_DISCARD_XPATH)
+        etree.strip_tags(subtree, 'a', 'ref', 'span')
         # extract content
         #for elem in subtree.xpath('.//*'):
         #    processed_elem = process_comments_node(elem, potential_tags)
@@ -436,7 +481,7 @@ def extract_comments(tree, dedupbool, config):
     return comments_body, temp_comments, len(temp_comments), tree
 
 
-def compare_extraction(tree, backup_tree, url, body, text, len_text, target_language, include_formatting, config):
+def compare_extraction(tree, backup_tree, url, body, text, len_text, target_language, include_formatting, include_links, include_images, config):
     '''Decide whether to choose own or external extraction
        based on a series of heuristics'''
     # bypass
@@ -457,7 +502,7 @@ def compare_extraction(tree, backup_tree, url, body, text, len_text, target_lang
         algo_flag = False
     elif len_algo > 2*len_text:
         algo_flag = True
-    elif not body.xpath('//p//text()') and len_algo > config.getint('DEFAULT', 'MIN_EXTRACTED_SIZE'):
+    elif not body.xpath('//p//text()') and len_algo > config.getint('DEFAULT', 'MIN_EXTRACTED_SIZE')*2:
         algo_flag = True  # borderline case
     else:
         LOGGER.debug('extraction values: %s %s for %s', len_text, len_algo, url)
@@ -476,7 +521,7 @@ def compare_extraction(tree, backup_tree, url, body, text, len_text, target_lang
             body, text, len_text = body2, text2, len_text2
         else:
             # post-processing: remove unwanted sections
-            body, text, len_text = sanitize_tree(body, include_formatting)
+            body, text, len_text = sanitize_tree(body, include_formatting, include_links, include_images)
     # try with justext
     elif len_text < config.getint('DEFAULT', 'MIN_EXTRACTED_SIZE'):
         LOGGER.error('not enough text %s', url)
@@ -484,15 +529,10 @@ def compare_extraction(tree, backup_tree, url, body, text, len_text, target_lang
         LOGGER.debug('justext length %s', len_text)
         if jt_result is False:
             # post-processing: remove unwanted sections
-            body, text, len_text = sanitize_tree(body, include_formatting)
+            body, text, len_text = sanitize_tree(body, include_formatting, include_links, include_images)
     else:
         if algo_flag is True:
-            body, text, len_text = sanitize_tree(body, include_formatting)
-    # second backup
-    #if len_text < MIN_EXTRACTED_SIZE:
-    #     body2, temp_text2, len_text2 = baseline(backup_tree)
-    #     if len_text2 > MIN_EXTRACTED_SIZE:
-    #         body, text, len_text = body2, len_text2, temp_text2
+            body, text, len_text = sanitize_tree(body, include_formatting, include_links, include_images)
     return body, text, len_text
 
 
@@ -507,7 +547,6 @@ def baseline(filecontent):
         the main text as string, and its length as integer.
 
     """
-
     tree = load_html(filecontent)
     postbody = etree.Element('body')
     if tree is None:
@@ -546,13 +585,13 @@ def baseline(filecontent):
     return postbody, temp_text, len(temp_text)
 
 
-def determine_returnstring(docmeta, output_format, tei_validation):
+def determine_returnstring(docmeta, output_format, include_formatting, include_links, tei_validation):
     '''Convert XML tree to chosen format, clean the result and output it as a string'''
     # XML (TEI) steps
     if 'xml' in output_format:
         # last cleaning
         for element in docmeta['body'].iter():
-            if element.tag != 'image' and len(element) == 0 and not element.text and not element.tail:
+            if element.tag != 'graphic' and len(element) == 0 and not element.text and not element.tail:
                 parent = element.getparent()
                 if parent is not None:
                     parent.remove(element)
@@ -566,41 +605,29 @@ def determine_returnstring(docmeta, output_format, tei_validation):
     # CSV, JSON and TXT output
     else:
         if output_format == 'csv':
-            posttext = xmltotxt(docmeta['body'])
+            posttext = xmltotxt(docmeta['body'], include_formatting, include_links)
             if docmeta['commentsbody'] is not None:
-                commentstext = xmltotxt(docmeta['commentsbody'])
+                commentstext = xmltotxt(docmeta['commentsbody'], include_formatting, include_links)
             else:
                 commentstext = ''
             returnstring = txttocsv(posttext, commentstext, docmeta)
         elif output_format == 'json':
             returnstring = build_json_output(docmeta)
         else:  # txt
-            returnstring = xmltotxt(docmeta['body'])
+            returnstring = xmltotxt(docmeta['body'], include_formatting, include_links)
             if docmeta['commentsbody'] is not None:
-                returnstring += '\n' + xmltotxt(docmeta['commentsbody'])
+                returnstring += '\n' + xmltotxt(docmeta['commentsbody'], include_formatting, include_links)
                 returnstring = returnstring.strip()
     return returnstring
 
 
-def map_format(output_format, csv_output, json_output, xml_output, tei_output):
-    '''Map existing options to format choice.'''
-    if output_format == 'txt' and any([csv_output, json_output, xml_output, tei_output]):
-        if csv_output is True:
-            output_format = 'csv'
-        elif json_output is True:
-            output_format = 'json'
-        elif xml_output is True:
-            output_format = 'xml'
-        elif tei_output is True:
-            output_format = 'xmltei'
-    return output_format
-
-
 def bare_extraction(filecontent, url=None, no_fallback=False,
                     include_comments=True, output_format='python', target_language=None,
-                    include_tables=True, include_images=False, include_formatting=False, deduplicate=False,
-                    date_extraction_params=None, with_metadata=False, max_tree_size=None,
-                    url_blacklist=None, config=DEFAULT_CONFIG):
+                    include_tables=True, include_images=False, include_formatting=False,
+                    include_links=False, deduplicate=False,
+                    date_extraction_params=None,
+                    only_with_metadata=False, with_metadata=False,
+                    max_tree_size=None, url_blacklist=None, config=DEFAULT_CONFIG):
     """Internal function for text extraction returning bare Python variables.
 
     Args:
@@ -610,15 +637,18 @@ def bare_extraction(filecontent, url=None, no_fallback=False,
         include_comments: Extract comments along with the main text.
         output_format: Define an output format, Python being the default
             and the interest of this internal function.
+            Other values: 'txt', 'csv', 'json', 'xml', or 'xmltei'.
         target_language: Define a language to discard invalid documents (ISO 639-1 format).
         include_tables: Take into account information within the HTML <table> element.
-        include_images: Take images into account.
+        include_images: Take images into account (experimental).
         include_formatting: Keep structural elements related to formatting
-            (only valuable if output_format is set to XML).
+            (present in XML format, converted to markdown otherwise).
+        include_links: Keep links along with their targets (experimental).
         deduplicate: Remove duplicate segments and documents.
         date_extraction_params: Provide extraction parameters to htmldate as dict().
-        with_metadata: Only keep documents featuring all essential metadata
+        only_with_metadata: Only keep documents featuring all essential metadata
             (date, title, url).
+        with_metadata: similar (will be deprecated).
         max_tree_size: Discard documents with too many elements.
         url_blacklist: Provide a blacklist of URLs as set() to filter out documents.
         config: Directly provide a configparser configuration.
@@ -633,6 +663,13 @@ def bare_extraction(filecontent, url=None, no_fallback=False,
     if url_blacklist is None:
         url_blacklist = set()
 
+    # deprecation warning
+    if with_metadata is True:
+        warnings.warn(
+            "with_metadata will be deprecated in a future version, use only_with_metadata instead",
+             PendingDeprecationWarning
+        )
+
     # load data
     try:
         tree = load_html(filecontent)
@@ -656,7 +693,7 @@ def bare_extraction(filecontent, url=None, no_fallback=False,
             if docmeta['url'] in url_blacklist:
                 raise ValueError
             # cut short if core elements are missing
-            if with_metadata is True and any(
+            if only_with_metadata is True or with_metadata is True and any(
                     x is None for x in
                     [docmeta['date'], docmeta['title'], docmeta['url']]
                 ):
@@ -668,28 +705,28 @@ def bare_extraction(filecontent, url=None, no_fallback=False,
         cleaned_tree = tree_cleaning(tree, include_tables, include_images)
 
         # convert tags, the rest does not work without conversion
-        cleaned_tree = convert_tags(cleaned_tree, include_formatting, include_tables, include_images)
+        cleaned_tree = convert_tags(cleaned_tree, include_formatting, include_tables, include_images, include_links)
 
         # comments first, then remove
         if include_comments is True:
             commentsbody, temp_comments, len_comments, cleaned_tree = extract_comments(cleaned_tree, deduplicate, config)
         else:
             commentsbody, temp_comments, len_comments = None, '', 0
+            #cleaned_tree = prune_unwanted_nodes(cleaned_tree, REMOVE_COMMENTS_XPATH)
 
         # extract content
-        postbody, temp_text, len_text, sure_thing = extract_content(cleaned_tree, include_tables, deduplicate, config)
+        postbody, temp_text, len_text, sure_thing = extract_content(cleaned_tree, include_tables, include_images, include_links, deduplicate, config)
 
         # compare if necessary
         if no_fallback is False:
-            #if sure_thing is False:
-            postbody, temp_text, len_text = compare_extraction(tree, backup_tree, url, postbody, temp_text, len_text, target_language, include_formatting, config)
+            postbody, temp_text, len_text = compare_extraction(tree, backup_tree, url, postbody, temp_text, len_text, target_language, include_formatting, include_links, include_images, config)
+            # add baseline as additional fallback
+            if len(postbody) == 0: # or len_text < config.getint('DEFAULT', 'MIN_EXTRACTED_SIZE'):
+                postbody, temp_text, len_text = baseline(filecontent)
         else:
             # rescue: try to use original/dirty tree
             if sure_thing is False and len_text < config.getint('DEFAULT', 'MIN_EXTRACTED_SIZE'):
                 postbody, temp_text, len_text = baseline(filecontent)
-                #tree = load_html(filecontent)
-                #tree = convert_tags(tree)
-                #postbody, temp_text, len_text, sure_thing = extract_content(tree)
                 LOGGER.debug('non-clean extracted length: %s (extraction)', len_text)
 
         # tree size sanity check
@@ -723,9 +760,9 @@ def bare_extraction(filecontent, url=None, no_fallback=False,
 
     # special case: python variables
     if output_format == 'python':
-        docmeta['text'] = xmltotxt(postbody)
+        docmeta['text'] = xmltotxt(postbody, include_formatting, include_links)
         if include_comments is True:
-            docmeta['comments'] = xmltotxt(commentsbody)
+            docmeta['comments'] = xmltotxt(commentsbody, include_formatting, include_links)
     else:
         docmeta['raw-text'], docmeta['body'], docmeta['commentsbody'] = temp_text, postbody, commentsbody
     return docmeta
@@ -733,10 +770,12 @@ def bare_extraction(filecontent, url=None, no_fallback=False,
 
 def extract(filecontent, url=None, record_id=None, no_fallback=False,
             include_comments=True, output_format='txt',
-            csv_output=False, json_output=False, xml_output=False, tei_output=False,
             tei_validation=False, target_language=None,
-            include_tables=True, include_images=False, include_formatting=False, deduplicate=False,
-            date_extraction_params=None, with_metadata=False, max_tree_size=None, url_blacklist=None,
+            include_tables=True, include_images=False, include_formatting=False,
+            include_links=False, deduplicate=False,
+            date_extraction_params=None,
+            only_with_metadata=False, with_metadata=False,
+            max_tree_size=None, url_blacklist=None,
             settingsfile=None, config=DEFAULT_CONFIG):
     """Main function exposed by the package:
        Wrapper for text extraction and conversion to chosen output format.
@@ -749,20 +788,18 @@ def extract(filecontent, url=None, record_id=None, no_fallback=False,
         include_comments: Extract comments along with the main text.
         output_format: Define an output format:
             'txt', 'csv', 'json', 'xml', or 'xmltei'.
-        csv_output: Set output format to CSV (alternative to output_format).
-        json_output: Set output format to JSON (alternative to output_format).
-        xml_output: Set output format to XML (alternative to output_format).
-        tei_output: Set output format to XML-TEI (alternative to output_format).
         tei_validation: Validate the XML-TEI output with respect to the TEI standard.
         target_language: Define a language to discard invalid documents (ISO 639-1 format).
         include_tables: Take into account information within the HTML <table> element.
-        include_images: Take images into account.
+        include_images: Take images into account (experimental).
         include_formatting: Keep structural elements related to formatting
             (only valuable if output_format is set to XML).
+        include_links: Keep links along with their targets (experimental).
         deduplicate: Remove duplicate segments and documents.
         date_extraction_params: Provide extraction parameters to htmldate as dict().
-        with_metadata: Only keep documents featuring all essential metadata
+        only_with_metadata: Only keep documents featuring all essential metadata
             (date, title, url).
+        with_metadata: similar (will be deprecated).
         max_tree_size: Discard documents with too many elements.
         url_blacklist: Provide a blacklist of URLs as set() to filter out documents.
         settingsfile: Use a configuration file to override the standard settings.
@@ -774,8 +811,6 @@ def extract(filecontent, url=None, record_id=None, no_fallback=False,
     """
     # configuration init
     config = use_config(settingsfile, config)
-    # metadata mapping for compatibility
-    output_format = map_format(output_format, csv_output, json_output, xml_output, tei_output)
     if url_blacklist is None:
         url_blacklist = set()
     # extraction
@@ -783,8 +818,10 @@ def extract(filecontent, url=None, record_id=None, no_fallback=False,
         filecontent, url=url, no_fallback=no_fallback,
         include_comments=include_comments, output_format=output_format,
         target_language=target_language, include_tables=include_tables, include_images=include_images,
-        include_formatting=include_formatting, deduplicate=deduplicate,
-        date_extraction_params=date_extraction_params, with_metadata=with_metadata,
+        include_formatting=include_formatting, include_links=include_links,
+        deduplicate=deduplicate,
+        date_extraction_params=date_extraction_params,
+        only_with_metadata=only_with_metadata, with_metadata=with_metadata,
         max_tree_size=max_tree_size, url_blacklist=url_blacklist, config=config,
         )
     if docmeta is None:
@@ -795,7 +832,7 @@ def extract(filecontent, url=None, record_id=None, no_fallback=False,
         # calculate fingerprint
         docmeta['fingerprint'] = content_fingerprint(docmeta['raw-text'])
     # return
-    return determine_returnstring(docmeta, output_format, tei_validation)
+    return determine_returnstring(docmeta, output_format, include_formatting, include_links, tei_validation)
 
 
 # for legacy and backwards compatibility
