@@ -27,6 +27,7 @@ from .xml import TEI_VALID_TAGS
 LOGGER = logging.getLogger(__name__)
 
 SANITIZED_XPATH = '//aside|//audio|//button|//fieldset|//figure|//footer|//iframe|//img|//image|//input|//label|//link|//nav|//noindex|//noscript|//object|//option|//select|//source|//svg|//time'
+SANITIZED_XPATH_WITH_IMAGES = '//aside|//audio|//button|//fieldset|//figure|//footer|//iframe|//input|//label|//link|//nav|//noindex|//noscript|//object|//option|//select|//source|//svg|//time'
 
 
 def jt_stoplist_init():
@@ -102,15 +103,25 @@ def justext_rescue(tree, url, target_language, postbody, len_text, text):
     return postbody, text, len_text, result_bool
 
 
-def sanitize_tree(tree, include_formatting=False):
+def sanitize_tree(tree, include_formatting=False, include_links=False, include_images=False):
     '''Convert and sanitize the output from the generic algorithm (post-processing)'''
+    if include_images is False:
+        sanitized_xpath = SANITIZED_XPATH
+    else:
+        sanitized_xpath = SANITIZED_XPATH_WITH_IMAGES
     # delete unnecessary elements
-    for elem in tree.xpath(SANITIZED_XPATH):
+    for elem in tree.xpath(sanitized_xpath):
         elem.getparent().remove(elem)
-    etree.strip_tags(tree, MANUALLY_STRIPPED + ['a', 'span'])
+    # elements to be stripped
+    stripped_list = MANUALLY_STRIPPED + ['a', 'span']
+    if include_links is True:
+        stripped_list.remove('a')
+    if include_images is True:
+        stripped_list.remove('img')
+    etree.strip_tags(tree, stripped_list)
     tree = prune_html(tree)
     # convert
-    cleaned_tree = convert_tags(tree, include_formatting)
+    cleaned_tree = convert_tags(tree, include_formatting=include_formatting, include_links=include_links, include_images=include_images)
     for elem in cleaned_tree.iter('td', 'th', 'tr'):
         # elem.text, elem.tail = trim(elem.text), trim(elem.tail)
         # finish table conversion
@@ -125,11 +136,6 @@ def sanitize_tree(tree, include_formatting=False):
     for tagname in [element.tag for element in set(cleaned_tree.iter())]:
         if tagname not in TEI_VALID_TAGS:
             sanitization_list.append(tagname)
-        #    if tagname in ('article', 'content', 'link', 'main', 'section', 'span'):
-        #        for element in cleaned_tree.iter(tagname):
-        #            merge_with_parent(element)
-        #    else:
-        #    print(tagname)
     etree.strip_tags(cleaned_tree, sanitization_list)
     text = trim(' '.join(cleaned_tree.itertext()))
     return cleaned_tree, text, len(text)
