@@ -35,30 +35,32 @@ HTML_CLEANER.remove_unknown_tags = False
 HTML_CLEANER.safe_attrs_only = False
 HTML_CLEANER.scripts = False
 HTML_CLEANER.style = False
-HTML_CLEANER.remove_tags = MANUALLY_STRIPPED
-HTML_CLEANER.kill_tags = MANUALLY_CLEANED
+#HTML_CLEANER.remove_tags = MANUALLY_STRIPPED
+#HTML_CLEANER.kill_tags = MANUALLY_CLEANED
 
 
 def tree_cleaning(tree, include_tables, include_images=False):
     '''Prune the tree by discarding unwanted elements'''
+    # determine cleaning strategy
+    cleaning_list, stripping_list = \
+        MANUALLY_CLEANED.copy(), MANUALLY_STRIPPED.copy()
     if include_tables is False:
-        MANUALLY_CLEANED.append('table')
+        cleaning_list.append('table')
     if include_images is True:
         # Many websites have <img> inside <figure> or <picture> or <source> tag
-        for element in ['figure', 'picture', 'source']:
-            MANUALLY_CLEANED.remove(element)
-        MANUALLY_STRIPPED.remove('img')
-    for expression in MANUALLY_CLEANED:
+        cleaning_list = [e for e in cleaning_list if e
+                         not in ('figure', 'picture', 'source')]
+        stripping_list.remove('img')
+    # delete targeted elements
+    for expression in cleaning_list:
         for element in tree.getiterator(expression):
             try:
-                element.drop_tree()
+                element.drop_tree() # faster when applicable
             except AttributeError:
                 element.getparent().remove(element)
+    HTML_CLEANER.kill_tags, HTML_CLEANER.remove_tags = cleaning_list, stripping_list
     # save space and processing time
-    tree = prune_html(tree)
-    tree = HTML_CLEANER.clean_html(tree)
-    # etree.strip_tags(tree, MANUALLY_STRIPPED)
-    return tree
+    return HTML_CLEANER.clean_html(prune_html(tree))
 
 
 def prune_html(tree):
@@ -72,18 +74,22 @@ def prune_html(tree):
     return tree
 
 
-def discard_unwanted(tree):
-    '''Delete unwanted sections'''
-    for expr in DISCARD_XPATH:
-        for subtree in tree.xpath(expr):
-            subtree.getparent().remove(subtree)
-    return tree
-
-
-def discard_unwanted_comments(tree):
-    '''delete unwanted comment sections'''
-    for expr in COMMENTS_DISCARD_XPATH:
+def prune_unwanted_nodes(tree, nodelist):
+    '''Prune the HTML tree by removing unwanted sections.'''
+    for expr in nodelist:
         for subtree in tree.xpath(expr):
+            # preserve tail text from deletion
+            if subtree.tail is not None:
+                previous = subtree.getprevious()
+                if previous is None:
+                    previous = subtree.getparent()
+                if previous is not None:
+                    # There is a previous node, append text to its tail
+                    if previous.tail is not None:
+                        previous.tail = ' '.join([previous.tail, subtree.tail])
+                    else:
+                        previous.tail = subtree.tail
+            # remove the node
             subtree.getparent().remove(subtree)
     return tree
 
@@ -98,6 +104,7 @@ def collect_link_info(links_xpath):
             continue
         linklen += subelemlen
         elemnum += 1
+        # TODO: unnecessary?
         if subelemlen < 10:
             shortelems += 1
         mylist.append(subelemtext)
@@ -106,16 +113,24 @@ def collect_link_info(links_xpath):
 
 def link_density_test(element):
     '''Remove sections which are rich in links (probably boilerplate)'''
-    links_xpath, mylist = element.xpath('.//link'), []
+    links_xpath, mylist = element.xpath('.//ref'), []
     if links_xpath:
-        elemtext = trim(element.text_content())
-        elemlen = len(elemtext)
-        if element.tag == 'p':
-            limitlen, threshold = 25, 0.9
+        elemtext = element.text_content()
+        elemlen = len(trim(elemtext))
+        #elemlen = len(trim(element.text_content()))
+        if element.tag == 'p': #  and not element.getparent().tag == 'item'
+            #if element.getnext() is None:
+            #    limitlen, threshold = 100, 0.8
+            #else:
+            limitlen, threshold = 25, 0.8
+            #if 'hi' in list(element):
+            #    limitlen, threshold = 100, 0.8
+        #elif element.tag == 'head':
+        #    limitlen, threshold = 50, 0.8
         else:
             if element.getnext() is None:
                 limitlen, threshold = 200, 0.66
-            #elif re.search(r'[.?!]', elemtext):
+            #elif re.search(r'[.?!:]', elemtext):
             #    limitlen, threshold = 150, 0.66
             else:
                 limitlen, threshold = 100, 0.66
@@ -123,8 +138,6 @@ def link_density_test(element):
             linklen, elemnum, shortelems, mylist = collect_link_info(links_xpath)
             if elemnum == 0:
                 return True, mylist
-            #if len(set(mylist))/len(mylist) <= 0.5:
-            #    return True, mylist
             LOGGER.debug('list link text/total: %s/%s – short elems/total: %s/%s', linklen, elemlen, shortelems, elemnum)
             if linklen >= threshold*elemlen or shortelems/elemnum >= threshold:
                 return True, mylist
@@ -136,11 +149,11 @@ def link_density_test_tables(element):
     '''Remove tables which are rich in links (probably boilerplate)'''
     # if element.getnext() is not None:
     #    return False
-    links_xpath = element.xpath('.//link')
+    links_xpath = element.xpath('.//ref')
     if links_xpath:
         elemlen = len(trim(element.text_content()))
         if elemlen > 250:
-            linklen, elemnum, shortelems, mylist = collect_link_info(links_xpath)
+            linklen, elemnum, shortelems, _ = collect_link_info(links_xpath)
             if elemnum == 0:
                 return True
             #if len(set(mylist))/len(mylist) <= 0.5:
@@ -149,12 +162,13 @@ def link_density_test_tables(element):
             if (elemlen < 1000 and linklen > 0.8*elemlen) or (elemlen > 1000 and linklen > 0.5*elemlen):
             #if linklen > 0.5 * elemlen:
                 return True
-            if shortelems > len(links_xpath) * 0.66:
-                return True
+            # does more harm than good (issue #76)
+            #if shortelems > len(links_xpath) * 0.66:
+            #    return True
     return False
 
 
-def convert_tags(tree, include_formatting=False, include_tables=False, include_images=False):
+def convert_tags(tree, include_formatting=False, include_tables=False, include_images=False, include_links=False):
     '''Simplify markup and convert relevant HTML tags to an XML standard'''
     # ul/ol → list / li → item
     for elem in tree.iter('ul', 'ol', 'dl'):
@@ -162,22 +176,33 @@ def convert_tags(tree, include_formatting=False, include_tables=False, include_i
         for subelem in elem.iter('dd', 'dt', 'li'):
             subelem.tag = 'item'
         for subelem in elem.iter('a'):
-            subelem.tag = 'link'
+            subelem.tag = 'ref'
     # divs
     for elem in tree.xpath('//div//a'):
-        elem.tag = 'link'
+        elem.tag = 'ref'
     # tables
     if include_tables is True:
         for elem in tree.xpath('//table//a'):
-            elem.tag = 'link'
+            elem.tag = 'ref'
     # images
     if include_images is True:
         for elem in tree.iter('img'):
-            elem.tag = 'image'
+            elem.tag = 'graphic'
     # delete links for faster processing
-    etree.strip_tags(tree, 'a')
+    if include_links is False:
+        etree.strip_tags(tree, 'a')
+    else:
+        for elem in tree.iter('a', 'ref'):
+            elem.tag = 'ref'
+            # replace href attribute and delete the rest
+            target = elem.get('href') # defaults to None
+            elem.attrib.clear()
+            if target is not None:
+                elem.set('target', target)
     # head tags + delete attributes
     for elem in tree.iter('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
+        elem.attrib.clear()
+        elem.set('rend', elem.tag)
         elem.tag = 'head'
     # br → lb
     for elem in tree.iter('br', 'hr'):
@@ -217,6 +242,11 @@ def convert_tags(tree, include_formatting=False, include_tables=False, include_i
     for elem in tree.iter('del', 's', 'strike'):
         elem.tag = 'del'
         elem.set('rend', 'overstrike')
+    # details + summary
+    for elem in tree.iter('details'):
+        elem.tag = 'div'
+        for subelem in elem.iter('summary'):
+            subelem.tag = 'head'
     return tree
 
 
