@@ -13,13 +13,13 @@ from htmldate import find_date
 from lxml import html
 
 from .metaxpaths import author_xpaths, categories_xpaths, tags_xpaths, title_xpaths
-from .utils import load_html, trim
+from .utils import line_processing, load_html, trim
 
 
 LOGGER = logging.getLogger(__name__)
 logging.getLogger('htmldate').setLevel(logging.WARNING)
 
-METADATA_LIST = ['title', 'author', 'url', 'hostname', 'description', 'sitename', 'date', 'categories', 'tags', 'fingerprint', 'id']
+METADATA_LIST = ['title', 'author', 'url', 'hostname', 'description', 'sitename', 'date', 'categories', 'tags', 'fingerprint', 'id', 'license']
 
 HTMLDATE_CONFIG = {'extensive_search': False, 'original_date': True}
 
@@ -32,6 +32,11 @@ JSON_NAME = re.compile(r'"@type":"[Aa]rticle", ?"name": ?"([^"\\]+)', re.DOTALL)
 JSON_HEADLINE = re.compile(r'"headline": ?"([^"\\]+)', re.DOTALL)
 URL_COMP_CHECK = re.compile(r'https?://|/')
 
+METANAME_AUTHOR = {'author', 'byl', 'dc.creator', 'dcterms.creator', 'sailthru.author'} # twitter:creator
+METANAME_TITLE = {'title', 'dc.title', 'dcterms.title', 'fb_title', 'sailthru.title', 'twitter:title'}
+METANAME_DESCRIPTION = {'description', 'dc.description', 'dcterms.description', 'dc:description', 'sailthru.description', 'twitter:description'}
+METANAME_PUBLISHER = {'copyright', 'dc.publisher', 'dcterms.publisher', 'publisher'}
+
 
 def normalize_json(inputstring):
     'Normalize unicode strings and trim the output'
@@ -71,7 +76,11 @@ def extract_json(tree, metadata):
         if '"publisher"' in elem.text:
             mymatch = JSON_PUBLISHER.search(elem.text)
             if mymatch and not ',' in mymatch.group(1):
-                metadata['sitename'] = normalize_json(mymatch.group(1))
+                candidate = normalize_json(mymatch.group(1))
+                if metadata['sitename'] is None or len(metadata['sitename']) < len(candidate):
+                    metadata['sitename'] = candidate
+                if metadata['sitename'].startswith('http') and not candidate.startswith('http'):
+                    metadata['sitename'] = candidate
         # category
         if '"articleSection"' in elem.text:
             mymatch = JSON_CATEGORY.search(elem.text)
@@ -134,7 +143,7 @@ def examine_meta(tree):
     if all((title, author, url, description, site_name)):  # if they are all defined
         metadata['title'], metadata['author'], metadata['url'], metadata['description'], metadata['sitename'] = title, author, url, description, site_name
         return metadata
-    tags = []
+    tags, backup_sitename = [], None
     # skim through meta tags
     for elem in tree.iterfind('.//head/meta[@content]'):
         # content
@@ -151,27 +160,24 @@ def examine_meta(tree):
             if elem.get('property') == 'article:tag':
                 tags.append(content_attr)
             elif elem.get('property') in ('author', 'article:author'):
-                if author is None:
-                    author = content_attr
+                author = author or content_attr
         # name attribute
         elif 'name' in elem.attrib:
             name_attr = elem.get('name').lower()
             # author
-            if name_attr in ('author', 'byl', 'dc.creator', 'dcterms.creator', 'sailthru.author'):  # twitter:creator
-                if author is None:
-                    author = content_attr
+            if name_attr in METANAME_AUTHOR:
+                author = author or content_attr
             # title
-            elif name_attr in ('title', 'dc.title', 'dcterms.title', 'fb_title', 'sailthru.title', 'twitter:title'):
-                if title is None:
-                    title = content_attr
+            elif name_attr in METANAME_TITLE:
+                title = title or content_attr
             # description
-            elif name_attr in ('description', 'dc.description', 'dcterms.description', 'dc:description', 'sailthru.description', 'twitter:description'):
-                if description is None:
-                    description = content_attr
+            elif name_attr in METANAME_DESCRIPTION:
+                description = description or content_attr
             # site name
-            elif name_attr in ('publisher', 'dc.publisher', 'dcterms.publisher', 'twitter:site', 'application-name') or 'twitter:app:name' in elem.get('name'):
-                if site_name is None:
-                    site_name = content_attr
+            elif name_attr in METANAME_PUBLISHER:
+                site_name = site_name or content_attr
+            elif name_attr in ('twitter:site', 'application-name') or 'twitter:app:name' in elem.get('name'):
+                backup_sitename = content_attr
             # url
             elif name_attr == 'twitter:url':
                 if url is None and validate_url(content_attr)[0] is True:
@@ -181,22 +187,23 @@ def examine_meta(tree):
                 tags.append(content_attr)
         elif 'itemprop' in elem.attrib:
             if elem.get('itemprop') == 'author':
-                if author is None:
-                    author = content_attr
+                author = author or content_attr
             elif elem.get('itemprop') == 'description':
-                if description is None:
-                    description = content_attr
+                description = description or content_attr
             elif elem.get('itemprop') == 'headline':
-                if title is None:
-                    title = content_attr
+                title = title or content_attr
             # to verify:
             #elif elem.get('itemprop') == 'name':
             #    if title is None:
             #        title = elem.get('content')
         # other types
         else:
-            if not 'charset' in elem.attrib and not 'http-equiv' in elem.attrib and not 'property' in elem.attrib:
-                LOGGER.debug(html.tostring(elem, pretty_print=False, encoding='unicode').strip())
+            if not any(key in elem.attrib for key in ('charset', 'http-equiv', 'property')):
+                LOGGER.debug('unknown attribute: %s', html.tostring(elem, pretty_print=False, encoding='unicode').strip())
+    # backups
+    if site_name is None and backup_sitename is not None:
+        site_name = backup_sitename
+    # copy
     metadata['title'], metadata['author'], metadata['url'], metadata['description'], metadata['sitename'], metadata['tags'] = title, author, url, description, site_name, tags
     return metadata
 
@@ -210,6 +217,7 @@ def extract_metainfo(tree, expressions, len_limit=200):
         for elem in tree.xpath(expression):
             content = elem.text_content()
             if content and len(content) < len_limit:
+                #LOGGER.debug('metadata found in: %s', expression)
                 return trim(content)
             i += 1
         if i > 1:
@@ -305,7 +313,7 @@ def extract_url(tree, default_url=None):
 def extract_sitename(tree):
     '''Extract the name of a site from the main title (if it exists)'''
     title_elem = tree.find('.//head/title')
-    if title_elem is not None:
+    if title_elem is not None and title_elem.text is not None:
         try:
             mymatch = re.search(r'^.*?[-|]\s+(.*)$', title_elem.text)
             if mymatch:
@@ -335,7 +343,8 @@ def extract_catstags(metatype, tree):
         element = tree.find('.//head//meta[@property="article:section"]')
         if element is not None and 'content' in element.attrib:
             results.append(element.attrib['content'])
-    return [trim(x) for x in results if x is not None]
+    results = [line_processing(x) for x in results if x is not None]
+    return [x for x in results if x is not None]
 
 
 def extract_metadata(filecontent, default_url=None, date_config=None):
@@ -381,8 +390,8 @@ def extract_metadata(filecontent, default_url=None, date_config=None):
     date_config['url'] = metadata['url']
     try:
         metadata['date'] = find_date(tree, **date_config)
-    # temporary fix for htmldate bug
-    except UnicodeError:
+    # temporary fixes for htmldate bugs # todo: remove later
+    except (TypeError, UnicodeError):
         pass
     # sitename
     if metadata['sitename'] is None:
@@ -410,9 +419,18 @@ def extract_metadata(filecontent, default_url=None, date_config=None):
     # tags
     if not metadata['tags']:
         metadata['tags'] = extract_catstags('tags', tree)
+    # license
+    for element in tree.xpath('//a[@rel="license"]',):
+        if element.text is not None:
+            metadata['license'] = trim(element.text)
+            break
     # for safety: length check
     for key, value in metadata.items():
         if value is not None and len(value) > 10000:
             metadata[key] = value[:9999] + '…'
+    # remove spaces and control characters
+    for item in metadata:
+        if metadata[item] is not None and isinstance(metadata[item], str):
+            metadata[item] = line_processing(metadata[item])
     # return
     return metadata
