@@ -8,11 +8,9 @@ Listing a series of settings that are applied module-wide.
 
 import configparser
 
-from multiprocessing import cpu_count
+from os import cpu_count
 from pathlib import Path
 
-from . import __version__
-
 
 def use_config(filename=None, config=None):
     'Use configuration object or read and parse a settings file'
@@ -31,9 +29,7 @@ DEFAULT_CONFIG = use_config()
 
 
 # Safety checks
-MAX_FILE_SIZE = 20000000
-MIN_FILE_SIZE = 10
-DOWNLOAD_THREADS = 5
+DOWNLOAD_THREADS = min(cpu_count(), 16)  # 16 processes at most
 TIMEOUT = 30
 LRU_SIZE = 4096
 
@@ -41,13 +37,9 @@ LRU_SIZE = 4096
 MAX_FILES_PER_DIRECTORY = 1000
 FILENAME_LEN = 8
 FILE_PROCESSING_CORES = min(cpu_count(), 16)  # 16 processes at most
-PROCESSING_TIMEOUT = 30
 
 # Network
-SLEEP_TIME = 2
-USER_AGENT = 'trafilatura/' + __version__ + ' (+https://github.com/adbar/trafilatura)'
-USER_AGENTS = [USER_AGENT]
-USER_AGENTS_NUM = len(USER_AGENTS)
+MAX_SITEMAPS_SEEN = 10000
 
 
 # filters
@@ -64,13 +56,13 @@ MANUALLY_CLEANED = [
     # other content
     'applet', 'audio', 'canvas', 'figure', 'map', 'picture', 'svg', 'video',
     # secondary
-    'area', 'blink', 'button', 'datalist', 'details', 'dialog',
+    'area', 'blink', 'button', 'datalist', 'dialog',
     'frame', 'frameset', 'fieldset', 'link', 'input', 'ins', 'label', 'legend',
     'marquee', 'math', 'menuitem', 'nav', 'noscript', 'optgroup', 'option',
     'output', 'param', 'progress', 'rp', 'rt', 'rtc', 'select', 'source',
-    'style', 'summary', 'track', 'template', 'textarea', 'time', 'use',
+    'style', 'track', 'template', 'textarea', 'time', 'use',
 ]
-# 'meta', 'hr', 'img', 'data'
+# 'meta', 'hr', 'img', 'data', 'details', 'summary'
 
 MANUALLY_STRIPPED = [
     'abbr', 'acronym', 'address', 'bdi', 'bdo', 'big', 'cite', 'data', 'dfn',
