@@ -12,7 +12,7 @@ import re
 import sys
 
 from functools import lru_cache
-from random import randint
+
 
 # CChardet is faster and can be more accurate
 try:
@@ -20,26 +20,15 @@ try:
 except ImportError:
     import chardet
 
+
 import urllib3
 
 from lxml import etree, html
 # from lxml.html.soupparser import fromstring as fromsoup
 
-from .settings import MAX_FILE_SIZE, MIN_FILE_SIZE, TIMEOUT, USER_AGENTS, USER_AGENTS_NUM
-
 
 LOGGER = logging.getLogger(__name__)
 
-# customize headers
-RETRY_STRATEGY = urllib3.util.Retry(
-    total=3,
-    redirect=2, # raise_on_redirect=False,
-    connect=0,
-    backoff_factor=TIMEOUT*2,
-    status_forcelist=[429, 500, 502, 503, 504],
-)
-urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
-HTTP_POOL = urllib3.PoolManager(retries=RETRY_STRATEGY)
 
 # collect_ids=False, default_doctype=False, huge_tree=True,
 HTML_PARSER = html.HTMLParser(remove_comments=True, remove_pis=True, encoding='utf-8')
@@ -63,9 +52,6 @@ NOPRINT_TRANS_TABLE = {
 # Regex to check image file extensions
 IMAGE_EXTENSION = re.compile(r'([^\s]+(\.(jpe?g|png|gif|bmp)))')
 
-# Regex for crude extraction of host/domain name
-HOSTINFO = re.compile(r'(https?://[^/]+)')
-
 
 def is_gz_file(contents):
     """Tell if a file's magic number corresponds to the GZip format"""
@@ -85,6 +71,7 @@ def isutf8(data):
 
 def detect_encoding(bytesobject):
     """Read the first chunk of input and return its encoding"""
+    #https://github.com/scrapy/w3lib/blob/master/w3lib/encoding.py
     # unicode-test
     if isutf8(bytesobject):
         return 'UTF-8'
@@ -92,7 +79,7 @@ def detect_encoding(bytesobject):
     guess = chardet.detect(bytesobject[:1999])
     LOGGER.debug('guessed encoding: %s, confidence: %s', guess['encoding'], guess['confidence'])
     # fallback on full response
-    if guess is None or guess['confidence'] < 0.95:
+    if guess is None or (guess['confidence'] is not None and guess['confidence'] < 0.95):
         guess = chardet.detect(bytesobject)
         LOGGER.debug('second-guessed encoding: %s, confidence: %s', guess['encoding'], guess['confidence'])
     return guess['encoding']
@@ -102,12 +89,18 @@ def decode_response(response):
     """Read the urllib3 object corresponding to the server response,
        check if it could be GZip and eventually decompress it, then
        try to guess its encoding and decode it to return a unicode string"""
+    # urllib3 response object / bytes switch
     if isinstance(response, bytes):
         resp_content = response
-    elif is_gz_file(response.data):
-        resp_content = gzip.decompress(response.data)
     else:
         resp_content = response.data
+    # decode GZipped data
+    if is_gz_file(resp_content):
+        try:
+            resp_content = gzip.decompress(resp_content)
+        except (EOFError, OSError):
+            logging.warning('invalid GZ file')
+    # detect encoding
     guessed_encoding = detect_encoding(resp_content)
     LOGGER.debug('response encoding: %s', guessed_encoding)
     # process
@@ -115,69 +108,30 @@ def decode_response(response):
     if guessed_encoding is not None:
         try:
             htmltext = resp_content.decode(guessed_encoding)
-        except UnicodeDecodeError:
-            LOGGER.warning('encoding error: %s', guessed_encoding)
+        except (LookupError, UnicodeDecodeError): # VISCII: lookup
+            LOGGER.warning('wrong encoding detected: %s', guessed_encoding)
+    else:
+        LOGGER.error('no encoding detected: %s', guessed_encoding)
     # force decoding # ascii instead?
     if htmltext is None:
         htmltext = str(resp_content, encoding='utf-8', errors='replace')
+        #try:
+        #    # frequent error
+        #    htmltext = resp_content.decode('cp1252').encode('utf-8')
+        #except UnicodeDecodeError:
+        #    htmltext = str(resp_content, encoding='utf-8', errors='replace')
     return htmltext
 
 
-def determine_headers():
-    'Decide on user-agent string'
-    if USER_AGENTS_NUM == 1:
-        rnumber = 0
-    else:
-        rnumber = randint(0, USER_AGENTS_NUM - 1)
-    headers = {
-    'User-Agent': USER_AGENTS[rnumber],
-    }
-    return headers
-
-
-def fetch_url(url, decode=True):
-    """Fetches page using urllib3 and decodes the response.
-
-    Args:
-        url: URL of the page to fetch.
-        decode: Decode response instead of returning Urllib3 response object (boolean).
-
-    Returns:
-        HTML code as string, or Urllib3 response object (headers + body), or empty string in case
-        the result is invalid, or None if there was a problem with the network.
-
-    """
-    # send
-    try:
-        # read by streaming chunks (stream=True, iter_content=xx)
-        # so we can stop downloading as soon as MAX_FILE_SIZE is reached
-        response = HTTP_POOL.request('GET', url, headers=determine_headers(),
-                                     timeout=TIMEOUT)
-    except urllib3.exceptions.NewConnectionError as err:
-        LOGGER.error('connection refused: %s %s', url, err)
-        return ''  # raise error instead?
-    except urllib3.exceptions.MaxRetryError as err:
-        LOGGER.error('retries/redirects: %s %s', url, err)
-        return ''  # raise error instead?
-    except urllib3.exceptions.TimeoutError as err:
-        LOGGER.error('connection timeout: %s %s', url, err)
-    except Exception as err:
-        logging.error('unknown error: %s %s', url, err) # sys.exc_info()[0]
-    else:
-        # safety checks
-        if response.status != 200:
-            LOGGER.error('not a 200 response: %s for URL %s', response.status, url)
-        elif response.data is None or len(response.data) < MIN_FILE_SIZE:
-            LOGGER.error('too small/incorrect for URL %s', url)
-            return ''  # raise error instead?
-        elif len(response.data) > MAX_FILE_SIZE:
-            LOGGER.error('too large: length %s for URL %s', len(response.data), url)
-            return ''  # raise error instead?
-        else:
-            if decode is True:
-                return decode_response(response.data)
-            return response
-    return None
+def is_dubious_html(htmlobject):
+    "Assess if the object is proper HTML (with a corresponding declaration)."
+    if isinstance(htmlobject, bytes):
+        if 'html' not in htmlobject[:50].decode(encoding='ascii', errors='ignore').lower():
+            return True
+    elif isinstance(htmlobject, str):
+        if 'html' not in htmlobject[:50].lower():
+            return True
+    return False
 
 
 def load_html(htmlobject):
@@ -189,11 +143,13 @@ def load_html(htmlobject):
         return htmlobject
     tree = None
     check_flag = False
+    # use urllib3 response directly
+    if isinstance(htmlobject, urllib3.response.HTTPResponse):
+        htmlobject = decode_response(htmlobject.data)
+    # sanity check
+    check_flag = is_dubious_html(htmlobject)
     # try to detect encoding and convert to string
     if isinstance(htmlobject, bytes):
-        # test
-        if 'html' not in htmlobject[:50].decode(encoding='ascii', errors='ignore').lower():
-            check_flag = True
         guessed_encoding = detect_encoding(htmlobject)
         if guessed_encoding is not None:
             if guessed_encoding == 'UTF-8':
@@ -209,9 +165,6 @@ def load_html(htmlobject):
             tree = html.fromstring(htmlobject, parser=RECOVERY_PARSER)
     # use string if applicable
     elif isinstance(htmlobject, str):
-        # test
-        if 'html' not in htmlobject[:50].lower():
-            check_flag = True
         try:
             tree = html.fromstring(htmlobject, parser=HTML_PARSER)
         except ValueError:
@@ -225,11 +178,10 @@ def load_html(htmlobject):
     # default to None
     else:
         LOGGER.error('this type cannot be processed: %s', type(htmlobject))
-    # further test: is it (well-formed) HTML at all?
-    if tree is not None and check_flag is True:
-        if len(tree) < 2:
-            LOGGER.error('parsed tree length: %s, wrong data type or not valid HTML', len(tree))
-            tree = None
+    # rejection test: is it (well-formed) HTML at all?
+    if tree is not None and check_flag is True and len(tree) < 2:
+        LOGGER.error('parsed tree length: %s, wrong data type or not valid HTML', len(tree))
+        tree = None
     #if tree is None:
     #    if isinstance(htmlobject, bytes) or isinstance(htmlobject, str):
     #        # more robust parsing
@@ -245,7 +197,7 @@ def txttocsv(text, comments, docmeta):
     if comments is not None:
         comments = trim(' '.join(comments.splitlines()))
     tsv_output = \
-        '{url}\t{fingerprint}\t{hostname}\t{doctitle}\t{docdate}\t{text}\t{comments}\n' \
+        '{url}\t{fingerprint}\t{hostname}\t{doctitle}\t{docdate}\t{text}\t{comments}\t{textlicense}\n' \
         .format(
         url=docmeta['url'],
         fingerprint=docmeta['fingerprint'],
@@ -253,7 +205,8 @@ def txttocsv(text, comments, docmeta):
         doctitle=docmeta['title'],
         docdate=docmeta['date'],
         text=text,
-        comments=comments
+        comments=comments,
+        textlicense=docmeta['license']
         )
     # add id up front if provided
     if docmeta['id'] is not None:
@@ -315,12 +268,14 @@ def is_image_file(imagesrc):
     return False
 
 
-def fix_relative_urls(baseurl, url):
-    'Prepend protocol and host information to relative links.'
-    if url.startswith('/'):
-        urlfix = baseurl + url
-    elif not url.startswith('http'):
-        urlfix = baseurl + '/' + url
-    else:
-        urlfix = url
-    return urlfix
+def filter_urls(linklist, urlfilter):
+    'Return a list of links corresponding to the given substring pattern.'
+    if urlfilter is None:
+        return sorted(set(linklist))
+    # filter links
+    newlist = [l for l in linklist if urlfilter in l]
+    # feedburner option
+    if len(newlist) == 0:
+        newlist = [l for l in linklist if urlfilter in l or 'feedburner' in l or 'feedproxy' in l]
+    return sorted(set(newlist))
+
