@@ -16,19 +16,22 @@ from pathlib import Path
 from lxml import etree
 
 from . import __version__
+from .filters import text_chars_test
 from .utils import sanitize
 
 
 LOGGER = logging.getLogger(__name__)
 # validation
 TEI_SCHEMA = str(Path(__file__).parent / 'data/tei-schema.pickle')
-TEI_VALID_TAGS = {'body', 'cell', 'code', 'del', 'div', 'fw', 'head', 'hi', 'item', \
-                  'lb', 'list', 'p', 'quote', 'row', 'table'}
-TEI_VALID_ATTRS = {'rend', 'rendition', 'role', 'type'}
+TEI_VALID_TAGS = {'body', 'cell', 'code', 'del', 'div', 'fw', 'graphic', 'head', 'hi', \
+                  'item', 'lb', 'list', 'p', 'quote', 'ref', 'row', 'table'}
+TEI_VALID_ATTRS = {'rend', 'rendition', 'role', 'target', 'type'}
 TEI_RELAXNG = None # to be downloaded later if necessary
 
 CONTROL_PARSER = etree.XMLParser(remove_blank_text=True)
 
+TEXTELEMS = {'code', 'fw', 'head', 'lb', 'list', 'p', 'quote', 'row', 'table'}
+
 
 def build_json_output(docmeta):
     '''Build JSON output based on extracted information'''
@@ -38,27 +41,37 @@ def build_json_output(docmeta):
     outputdict['excerpt'] = outputdict.pop('description')
     outputdict['categories'] = ';'.join(outputdict['categories'])
     outputdict['tags'] = ';'.join(outputdict['tags'])
-    outputdict['text'] = xmltotxt(outputdict.pop('body'))
+    outputdict['text'] = xmltotxt(outputdict.pop('body'), include_formatting=False, include_links=False)
     if outputdict['commentsbody'] is not None:
-        outputdict['comments'] = xmltotxt(outputdict.pop('commentsbody'))
+        outputdict['comments'] = xmltotxt(outputdict.pop('commentsbody'), include_formatting=False, include_links=False)
     else:
         del outputdict['commentsbody']
-    return json.dumps(outputdict)
+    return json.dumps(outputdict, ensure_ascii=False)
 
 
 def clean_attributes(tree):
-    '''Remove unnecessary attributes'''
+    '''Remove unnecessary attributes.'''
     for elem in tree.iter():
-        if elem.tag not in ('del', 'hi', 'image'):
+        if elem.tag not in ('del', 'fw', 'graphic', 'head', 'hi', 'ref'):
             elem.attrib.clear()
     return tree
 
 
+def remove_empty_elements(tree):
+    '''Remove text elements without text.'''
+    for element in tree.iter('fw', 'head', 'hi', 'item', 'p'):
+        if len(element) == 0 and text_chars_test(element.text) is False and text_chars_test(element.tail) is False:
+            element.getparent().remove(element)
+    return tree
+
+
 def build_xml_output(docmeta):
     '''Build XML output tree based on extracted information'''
     output = etree.Element('doc')
     output = add_xml_meta(output, docmeta)
     docmeta['body'].tag = 'main'
+    # clean XML tree
+    docmeta['body'] = remove_empty_elements(docmeta['body'])
     output.append(clean_attributes(docmeta['body']))
     if docmeta['commentsbody'] is not None:
         docmeta['commentsbody'].tag = 'comments'
@@ -102,6 +115,8 @@ def add_xml_meta(output, docmeta):
             output.set('categories', ';'.join(docmeta['categories']))
         if docmeta['tags'] is not None:
             output.set('tags', ';'.join(docmeta['tags']))
+        if docmeta['license'] is not None:
+            output.set('license', docmeta['license'])
         if docmeta['id'] is not None:
             output.set('id', docmeta['id'])
         if docmeta['fingerprint'] is not None:
@@ -159,9 +174,30 @@ def validate_tei(tei):  # , filename=""
     return result
 
 
-def replace_element_text(element):
+def replace_element_text(element, include_formatting, include_links):
     '''Determine element text based on text and tail'''
     full_text = ''
+    # handle formatting
+    if include_formatting is True and element.text is not None:
+        if element.tag == 'hi':
+            if element.get('rend') == '#b':
+                element.text = ''.join(['**', element.text, '**'])
+            elif element.get('rend') == '#i':
+                element.text = ''.join(['*', element.text, '*'])
+            elif element.get('rend') == '#u':
+                element.text = ''.join(['__', element.text, '__'])
+            elif element.get('rend') == '#t':
+                element.text = ''.join(['`', element.text, '`'])
+        elif element.tag == 'del':
+            element.text = ''.join(['~~', element.text, '~~'])
+    # handle links
+    if include_links is True and element.tag == 'ref':
+        try:
+            element.text = ''.join(['[', element.text, ']', '(', element.get('target'), ')'])
+        except TypeError:
+            LOGGER.warning('missing link attribute: %s %s', element.text, element.attrib)
+            element.text = ''.join(['[', element.text, ']'])
+    # handle text
     if element.text is not None and element.tail is not None:
         full_text = ' '.join([element.text, element.tail])
     elif element.text is not None and element.tail is None:
@@ -171,12 +207,12 @@ def replace_element_text(element):
     return full_text
 
 
-def merge_with_parent(element):
-    '''Merge element with its parent'''
+def merge_with_parent(element, include_formatting=False, include_links=False):
+    '''Merge element with its parent and convert formatting to markdown.'''
     parent = element.getparent()
     if parent is None:
         return
-    full_text = replace_element_text(element)
+    full_text = replace_element_text(element, include_formatting, include_links)
     previous = element.getprevious()
     if previous is not None:
         # There is a previous node, append text to its tail
@@ -193,24 +229,30 @@ def merge_with_parent(element):
     parent.remove(element)
 
 
-def xmltotxt(xmloutput):
-    '''Convert to plain text format'''
+def xmltotxt(xmloutput, include_formatting, include_links):
+    '''Convert to plain text format and optionally preserve formatting as markdown.'''
     returnlist = []
     # etree.strip_tags(xmloutput, 'div', 'main', 'span')
     # remove and insert into the previous tag
-    for element in xmloutput.xpath('//hi|//link'):
-        merge_with_parent(element)
+    for element in xmloutput.xpath('//hi|//ref'):
+        merge_with_parent(element, include_formatting, include_links)
         continue
     # iterate and convert to list of strings
     for element in xmloutput.iter():
         # process text
         if element.text is None and element.tail is None:
+            if element.tag == 'graphic':
+                returnlist.extend(['\n', element.get('src')])
+                if element.get('alt') is not None:
+                    returnlist.extend([' ', element.get('alt')])
+                if element.get('title') is not None:
+                    returnlist.extend([' ', element.get('title')])
             # newlines for textless elements
-            if element.tag in ('row', 'table'):
+            if element.tag in ('graphic', 'row', 'table'):
                 returnlist.append('\n')
             continue
-        textelement = replace_element_text(element)
-        if element.tag in ('code', 'fw', 'head', 'lb', 'list', 'p', 'quote', 'row', 'table'):
+        textelement = replace_element_text(element, include_formatting, include_links)
+        if element.tag in TEXTELEMS:
             returnlist.extend(['\n', textelement, '\n'])
         elif element.tag == 'item':
             returnlist.extend(['\n- ', textelement, '\n'])
@@ -219,7 +261,7 @@ def xmltotxt(xmloutput):
         elif element.tag == 'comments':
             returnlist.append('\n\n')
         else:
-            # print(element.tag)
+            LOGGER.debug('unexpected element: %s', element.tag)
             returnlist.extend([textelement, ' '])
     return sanitize(''.join(returnlist))
 
@@ -258,8 +300,14 @@ def write_fullheader(header, docmeta):
         bib_author = etree.SubElement(bib_titlestmt, 'author')
         bib_author.text = docmeta['author']
     publicationstmt_a = etree.SubElement(filedesc, 'publicationStmt')
+    # license, if applicable
+    if docmeta['license'] is not None:
+        availability = etree.SubElement(publicationstmt_a, 'availability')
+        avail_p = etree.SubElement(availability, 'p')
+        avail_p.text = docmeta['license']
     # insert an empty paragraph for conformity
-    publicationstmt_p = etree.SubElement(publicationstmt_a, 'p')
+    else:
+        publicationstmt_p = etree.SubElement(publicationstmt_a, 'p')
     notesstmt = etree.SubElement(filedesc, 'notesStmt')
     if docmeta['id'] is not None:
         idno = etree.SubElement(notesstmt, 'note', type='id')
@@ -268,6 +316,7 @@ def write_fullheader(header, docmeta):
     fingerprint.text = docmeta['fingerprint']
     sourcedesc = etree.SubElement(filedesc, 'sourceDesc')
     source_bibl = etree.SubElement(sourcedesc, 'bibl')
+    # determination of sigle string
     if docmeta['sitename'] and docmeta['date']:
         sigle = docmeta['sitename'] + ', ' + docmeta['date']
     elif not docmeta['sitename'] and docmeta['date']:
@@ -292,7 +341,7 @@ def write_fullheader(header, docmeta):
     publicationstmt = etree.SubElement(biblfull, 'publicationStmt')
     publication_publisher = etree.SubElement(publicationstmt, 'publisher')
     if docmeta['hostname'] is not None:
-        publisherstring = docmeta['sitename'] + '(' + docmeta['hostname'] + ')'
+        publisherstring = docmeta['sitename'].strip() + ' (' + docmeta['hostname'] + ')'
     else:
         publisherstring = docmeta['sitename']
     publication_publisher.text = publisherstring
